// $Id: TclScript.cpp,v 1.2 2001/10/13 18:50:30 Administrator Exp $
#include <activscp.h>
#include "typeinfo.h"
#include "refCmd.h"

#define NAMESPACE "::TclScriptEngine::"
#define PACKAGE_NAME "TclScript"
#define PACKAGE_VERSION "1.0"

#ifdef __BORLANDC__
#undef DLLEXPORT
#define DLLEXPORT __declspec(dllexport)
#endif

int
getiteminfoCmd (
    ClientData,
    Tcl_Interp *interp,
    int objc,
    Tcl_Obj *CONST objv[])
{
    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "handle name");
	return TCL_ERROR;
    }

    Reference *pRef = referenceHandleSupport.find(objv[1]);
    if (pRef == 0) {
        const char *arg = Tcl_GetStringFromObj(objv[1], 0);
        Tcl_AppendResult(interp, "invalid handle ", arg, NULL);
	return TCL_ERROR;
    }

    try {
        HRESULT hr;

        IActiveScriptSitePtr pScriptSite;
        hr = pRef->unknown()->QueryInterface(
            IID_IActiveScriptSite, reinterpret_cast<void **>(&pScriptSite));
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        TclObject name(objv[2]);
        IUnknownPtr pUnknown;
        hr = pScriptSite->GetItemInfo(
            name.getUnicode(), SCRIPTINFO_IUNKNOWN, &pUnknown, 0);
        if (FAILED(hr)) {
            _com_issue_error(hr);
        }

        Reference *pNewRef = Reference::newReference(pUnknown);
        Tcl_SetObjResult(
            interp, referenceHandleSupport.newObj(interp, pNewRef));
    }
    catch (_com_error &e) {
        return setComErrorResult(interp, e, __FILE__, __LINE__);
    }

    return TCL_OK;
}

#if 0
// {4964A490-CB79-11D2-A252-00104BD35090}
DEFINE_GUID(IID_IHost2, 0x4964A490, 0xCB79, 0x11D2, 0xA2, 0x52, 0x0,
         0x10, 0x4B, 0xD3, 0x50, 0x90);

int
patchWScriptEchoParameterCmd (
    ClientData,
    Tcl_Interp *interp,
    int objc,
    Tcl_Obj *CONST objv[])
{
    Interface *pInterface = InterfaceManager::instance().find(IID_IHost2);
    return TCL_OK;
}
#endif

/*
 *	This procedure performs application-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the startup script.
 */
extern "C" DLLEXPORT int
Tclscript_Init (Tcl_Interp *interp)
{
#if TCL_MINOR_VERSION >= 1
    // Stubs were introduced in Tcl 8.1.
    if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
        return TCL_ERROR;
    }
#endif

    Tcl_CreateObjCommand(interp, NAMESPACE "getiteminfo", getiteminfoCmd, 0, 0);

    return Tcl_PkgProvide(interp, PACKAGE_NAME, PACKAGE_VERSION);
}

/*
 *	This procedure initializes commands for a safe interpreter.
 *	You would leave out of this procedure any commands you deemed unsafe.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 */
extern "C" DLLEXPORT int
Tclscript_SafeInit (Tcl_Interp *interp)
{
#if TCL_MINOR_VERSION >= 1
    // Stubs were introduced in Tcl 8.1.
    if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
        return TCL_ERROR;
    }
#endif

    return Tcl_PkgProvide(interp, PACKAGE_NAME, PACKAGE_VERSION);
}
